import path from 'path';
import fs from 'fs';
import mkdirp from 'mkdirp';
import chokidar from 'chokidar';
import consola from 'consola';
import express from 'express';
import { GraphQLClient } from 'graphql-request';
import { generate } from '@graphql-codegen/cli';
import * as PluginTypescript from '@graphql-codegen/typescript';
import * as PluginTypescriptOperations from '@graphql-codegen/typescript-operations';
import * as PluginSchemaAst from '@graphql-codegen/schema-ast';

// -- Unbuild CommonJS Shims --
import __cjs_url__ from 'url';
import __cjs_path__ from 'path';
import __cjs_mod__ from 'module';
const __filename = __cjs_url__.fileURLToPath(import.meta.url);
const __dirname = __cjs_path__.dirname(__filename);
const require = __cjs_mod__.createRequire(import.meta.url);


function getVariables(vars) {
  try {
    return JSON.parse(vars);
  } catch (error) {
    return {};
  }
}
function buildHeaders(req, name, type, config) {
  if (config?.buildHeaders) {
    return config.buildHeaders(req, name, type);
  }
  if (config?.fetchOptions?.headers) {
    return config.fetchOptions.headers;
  }
  return {};
}
function createServerMiddleware(graphqlServer, queries, mutations, config) {
  const app = express();
  app.use(express.json());
  const clients = new Map();
  function getClient(endpoint) {
    if (!clients.has(endpoint)) {
      const client = new GraphQLClient(endpoint);
      clients.set(endpoint, client);
    }
    return clients.get(endpoint);
  }
  function getEndpoint(req) {
    if (config?.buildEndpoint) {
      return config.buildEndpoint(req);
    }
    return graphqlServer;
  }
  if (config?.middleware) {
    app.use(config.middleware);
  }
  async function query(req, res) {
    const name = req.query.name;
    if (!name || !queries.has(name)) {
      res.status(404).send();
      return;
    }
    try {
      const headers = buildHeaders(req, name, "query", config);
      const variables = getVariables(req.query.variables);
      const query2 = queries.get(name);
      const endpoint = getEndpoint(req);
      const client = getClient(endpoint);
      const response = await client.rawRequest(query2, variables, headers);
      if (config?.onQueryResponse) {
        return config.onQueryResponse(response, req, res);
      }
      return res.json(response.data);
    } catch (e) {
      if (config?.onQueryError) {
        return config.onQueryError(e, req, res);
      }
      return res.status(500).send();
    }
  }
  async function mutate(req, res) {
    const name = req.query.name;
    if (!name || !mutations.has(name)) {
      res.status(404).send();
      return;
    }
    const mutation = mutations.get(name);
    try {
      const headers = buildHeaders(req, name, "mutation", config);
      const endpoint = getEndpoint(req);
      const client = getClient(endpoint);
      const response = await client.request(mutation, req.body, headers);
      if (config?.onMutationResponse) {
        return config.onMutationResponse(response, req, res);
      }
      return res.json(response);
    } catch (error) {
      if (config?.onMutationError) {
        return config.onMutationError(error, req, res);
      }
      return res.status(500).send();
    }
  }
  app.get("/query", query);
  app.post("/mutate", mutate);
  return app;
}

const fragmentImport = require("@graphql-fragment-import/lib/inline-imports");
function graphqlImport(path, resolver) {
  return fragmentImport(path, {
    resolveImport(identifier) {
      return resolver(identifier);
    },
    resolveOptions: {
      basedir: "./"
    }
  });
}

const typescriptConfig = {
  exportFragmentSpreadSubTypes: true,
  preResolveTypes: true,
  skipTypeNameForRoot: true
};
function pluginLoader(name) {
  if (name === "@graphql-codegen/typescript") {
    return Promise.resolve(PluginTypescript);
  } else if (name === "@graphql-codegen/typescript-operations") {
    return Promise.resolve(PluginTypescriptOperations);
  } else {
    return Promise.resolve(PluginSchemaAst);
  }
}
function codegen(graphqlServer, options) {
  const schemaPath = path.resolve(options.schemaOutputPath, "schema.graphql");
  function generateSchema() {
    const schema = options.skipSchemaDownload ? schemaPath : { [graphqlServer]: options.schemaOptions };
    const configSchemaAst = { ...typescriptConfig, sort: true };
    return generate({
      schema,
      pluginLoader,
      generates: {
        [schemaPath]: {
          plugins: [{ "schema-ast": configSchemaAst }],
          config: configSchemaAst
        },
        [path.resolve(options.typesOutputPath, "graphql-schema.d.ts")]: {
          plugins: [{ typescript: typescriptConfig }],
          config: typescriptConfig
        }
      }
    }, true);
  }
  function generateTypes() {
    const config = {
      ...typescriptConfig,
      onlyOperationTypes: true
    };
    return generate({
      schema: schemaPath,
      pluginLoader,
      documents: path.resolve(options.resolvedQueriesPath, "./*.graphql"),
      generates: {
        [path.resolve(options.typesOutputPath, "graphql-operations.d.ts")]: {
          plugins: ["typescript", { "typescript-operations": config }],
          config
        }
      }
    }, true);
  }
  return { generateSchema, generateTypes };
}

const logger = consola.withTag("nuxt-graphql-middleware");
const PLUGIN_PATH = path.resolve(__dirname, "../dist/plugin.mjs");
var FileType;
(function(FileType2) {
  FileType2["Query"] = "query";
  FileType2["Mutation"] = "mutation";
})(FileType || (FileType = {}));
function resolveGraphqlFile(file, resolver) {
  return fs.promises.readFile(file).then((buffer) => buffer.toString()).then((source) => graphqlImport(source, resolver));
}
function writeSource(dest, type, name, source) {
  const fileName = `${type}.${name}.graphql`;
  const out = path.resolve(dest, fileName);
  return fs.promises.writeFile(out, source);
}
function resolveGraphql(files, map, resolver, filesMap, type, outputPath) {
  return Promise.all(Object.keys(files).map((name) => {
    const filePath = files[name];
    const file = resolver(filePath);
    return resolveGraphqlFile(file, resolver).then((source) => {
      map.set(name, source);
      if (outputPath) {
        writeSource(outputPath, type, name, source);
      }
      filesMap.set(file, {
        type,
        name,
        file: filePath
      });
    });
  }));
}
const graphqlMiddleware = async function() {
  const resolver = this.nuxt.resolver.resolveAlias;
  const options = this.options;
  const PORT = this.options?.server?.port || 3e3;
  const provided = this.options.graphqlMiddleware || {};
  const config = {
    graphqlServer: provided.graphqlServer || "",
    typescript: {
      enabled: !!provided.typescript?.enabled,
      schemaOptions: provided.typescript?.schemaOptions,
      resolvedQueriesPath: provided.typescript?.resolvedQueriesPath || provided.outputPath || "",
      schemaOutputPath: provided.typescript?.schemaOutputPath || "~/schema",
      typesOutputPath: provided.typescript?.typesOutputPath || "~/types",
      skipSchemaDownload: !!provided.typescript?.skipSchemaDownload
    },
    endpointNamespace: provided.endpointNamespace || "/__graphql_middleware",
    debug: provided.debug || options.dev,
    queries: provided.queries || {},
    mutations: provided.mutations || {},
    outputPath: provided.outputPath || "",
    server: provided.server,
    plugin: {
      enabled: !!provided.plugin?.enabled,
      port: 4e3,
      cacheInBrowser: !!provided.plugin?.cacheInBrowser,
      cacheInServer: !!provided.plugin?.cacheInServer
    }
  };
  if (config.plugin?.enabled) {
    this.addPlugin({
      filename: "graphqlMiddleware.js",
      src: PLUGIN_PATH,
      options: {
        namespace: config.endpointNamespace,
        port: PORT,
        cacheInBrowser: config.plugin?.cacheInBrowser ? "true" : "false",
        cacheInServer: config.plugin?.cacheInServer ? "true" : "false"
      }
    });
  }
  const fileMap = new Map();
  const queries = new Map();
  const mutations = new Map();
  const outputPath = config.outputPath ? resolver(config.outputPath) : "";
  await mkdirp(outputPath);
  const schemaOutputPath = resolver(config.typescript?.schemaOutputPath);
  const typesOutputPath = resolver(config.typescript?.typesOutputPath);
  const { generateSchema, generateTypes } = codegen(config.graphqlServer, {
    resolvedQueriesPath: config.outputPath,
    schemaOptions: config.typescript?.schemaOptions,
    skipSchemaDownload: config.typescript?.skipSchemaDownload,
    schemaOutputPath,
    typesOutputPath
  });
  if (config.typescript?.enabled) {
    if (!outputPath) {
      throw new Error("TypeScript enabled, but no outputPath given.");
    }
    await mkdirp(schemaOutputPath);
    await generateSchema();
  }
  function build() {
    logger.log("Building GraphQL files...");
    return Promise.all([
      resolveGraphql(config.queries, queries, resolver, fileMap, FileType.Query, outputPath),
      resolveGraphql(config.mutations, mutations, resolver, fileMap, FileType.Mutation, outputPath)
    ]).then(() => {
      logger.success("Finished building GraphQL files");
      if (config.typescript?.enabled) {
        return generateTypes().then(() => {
          logger.success("Finished generating GraphQL TypeScript files.");
        });
      }
    });
  }
  function watchFiles() {
    const ignored = ["node_modules", ".nuxt"];
    if (config.outputPath) {
      ignored.push(config.outputPath);
    }
    const filesWatcher = chokidar.watch("./**/*.graphql", {
      ignoreInitial: true,
      ignored
    });
    if (filesWatcher) {
      logger.info("Watching for query changes");
      filesWatcher.on("change", () => {
        build();
      });
    }
    return filesWatcher;
  }
  let watcher;
  if (this.nuxt.options.dev) {
    this.nuxt.hook("build:done", () => {
      watcher = watchFiles();
    });
    this.nuxt.hook("close", () => {
      if (watcher) {
        watcher.close();
        watcher = void 0;
      }
    });
  }
  build().then(() => {
    if (options.debug) {
      logger.info("Available queries and mutations:");
      console.table(Array.from(fileMap.entries()).map(([_key, value]) => value));
    }
  });
  this.addServerMiddleware({
    path: config.endpointNamespace,
    handler: createServerMiddleware(config.graphqlServer, queries, mutations, config.server)
  });
};

export { graphqlMiddleware as default };
